@@ -1,10 +1,10 @@
 ##----------------------------------------------------------------------------
 ## Data Dump Beautifier - ~/lib/Data/Pretty.pm
-## Version v0.1.2
+## Version v0.1.3
 ## Copyright(c) 2023 DEGUEST Pte. Ltd.
 ## Author: Jacques Deguest <jack@deguest.jp>
 ## Created 2023/08/06
-## Modified 2023/08/07
+## Modified 2023/08/08
 ## All rights reserved
 ## 
 ## 
@@ -19,7 +19,7 @@ BEGIN
     use vars qw(
         @EXPORT @EXPORT_OK $VERSION $DEBUG
         %seen %refcnt @dump @fixup %require
-        $TRY_BASE64 @FILTERS $INDENT $LINEWIDTH $SHOW_UTF8
+        $TRY_BASE64 @FILTERS $INDENT $LINEWIDTH $SHOW_UTF8 $CODE_DEPARSE
     );
     use subs qq(dump);
     use overload ();
@@ -28,16 +28,17 @@ BEGIN
     @EXPORT = qw( dd ddx );
     @EXPORT_OK = qw( dump pp dumpf quote );
     our $DEBUG = 0;
-    our $VERSION = 'v0.1.2';
+    our $VERSION = 'v0.1.3';
 };
 
 use strict;
 use warnings;
 
 $TRY_BASE64 = 50 unless defined $TRY_BASE64;
-$INDENT = "  " unless defined $INDENT;
+$INDENT = '    ' unless defined $INDENT;
 $LINEWIDTH = 60 unless defined $LINEWIDTH;
 $SHOW_UTF8 = 1 unless defined $SHOW_UTF8;
+$CODE_DEPARSE = 1 unless defined $CODE_DEPARSE;
 
 {
     no warnings 'once';
@@ -68,8 +69,16 @@ sub dump
     my $name = "a";
     my @dump;
 
+    my $use_qw = &_use_qw( [@_] );
     for my $v (@_) {
-        my $val = _dump($v, $name, [], tied($v));
+        # my $val = _dump($v, $name, [], tied($v));
+        my $val = _dump(
+            $v,
+            name => $name,
+            idx => [],
+            dont_remember => tied($v),
+            use_qw => $use_qw,
+        );
         push(@dump, [$name, $val]);
     } continue {
         $name++;
@@ -96,12 +105,16 @@ sub dump
     }
 
     my $paren = (@dump != 1);
-    $out .= "(" if $paren;
-    $out .= format_list($paren, undef,
-			map {defined($_->[1]) ? $_->[1] : "\$".$_->[0]}
-			    @dump
-		       );
-    $out .= ")" if $paren;
+    my $formatted = format_list(
+        paren => $paren,
+        comment => undef,
+        values => [map {defined($_->[1]) ? $_->[1] : "\$".$_->[0]} @dump],
+        use_qw => $use_qw,
+    );
+    my $has_qw = substr( $formatted, 0, 2 ) eq 'qw';
+    $out .= "(" if( $paren && !$has_qw );
+    $out .= $formatted;
+    $out .= ")" if( $paren && !$has_qw );
 
     if (%refcnt || %require) {
         $out .= ";\n";
@@ -120,19 +133,21 @@ sub dumpf {
 
 sub format_list
 {
-    my $paren = shift;
-    my $comment = shift;
+    my $opts = {@_};
+    my $paren = $opts->{paren};
+    my $comment = $opts->{comment};
     my $indent_lim = $paren ? 0 : 1;
-    my $use_qw = !$paren;
+    my $use_qw = defined( $opts->{use_qw} ) ? $opts->{use_qw} : 1;
+    my $values = $opts->{values};
     
-    if (@_ > 3) {
+    if (@$values > 3) {
         # my $use_quotes = 0;
         # can we use range operator to shorten the list?
         my $i = 0;
-        while ($i < @_) {
+        while ($i < @$values) {
             my $j = $i + 1;
-            my $v = $_[$i];
-            while ($j < @_) {
+            my $v = $values->[$i];
+            while ($j < @$values) {
                 # NOTE: allow string increment too?
                 if ($v eq "0" || $v =~ /^-?[1-9]\d{0,9}\z/) {
                     $v++;
@@ -145,71 +160,63 @@ sub format_list
                 else {
                     last;
                 }
-                last if $_[$j] ne $v;
+                last if $values->[$j] ne $v;
                 $j++;
             }
             if ($j - $i > 3) {
-                splice(@_, $i, $j - $i, "$_[$i] .. $_[$j-1]");
-                # splice(@_, $i, $j - $i, ( $use_quotes ? "\"$_[$i]\" .. \"$_[$j-1]\"" : "$_[$i] .. $_[$j-1]" ) );
+                splice(@$values, $i, $j - $i, "$values->[$i] .. $values->[$j-1]");
                 $use_qw = 0;
             }
             $i++;
         }
     }
 
-    unless( !$use_qw )
+    if( $use_qw )
     {
-        # Quick check if we are dealing with a simple array of words/terms
-        # and thus if we can use qw( .... ) instead of ( "some", "thing", "else" )
-        my $only_numbers = 0;
         my @repl;
-        foreach my $v ( @_ )
+        foreach my $v ( @$values )
         {
-            if( !defined( $v ) || 
-                ref( $v ) || 
-                # See perlop/"qw/STRING/" section
-                ( !ref( $v ) && ( $v =~ /[\,\#\$\\[:blank:]\h\v]/ || $v eq '[]' || $v eq '{}' || $v eq 'undef' ) ) )
-            {
-                $use_qw = 0;
-                @repl = ();
-                last;
-            }
             ( my $v2 = $v ) =~ s/^\"|\"$//g;
             push( @repl, $v2 );
-            $only_numbers++ if( $v2 =~ /^[-+]?(?:0|[1-9]\d*)(?:\.\d+)?\z/ );
         }
-        # Don't use qw() if we are only dealing with numbers; or
-        # if there is only one value.
-        $use_qw = 0 if( $only_numbers == scalar( @_ ) || scalar( @_ ) == 1 );
-        @_ = @repl if( $use_qw );
+        @$values = @repl;
     }
 
-    my $tmp = "@_";
-#     if ($comment || (@_ > $indent_lim && (length($tmp) > $LINEWIDTH || $tmp =~ /\n/))) {
-#         my @elem = @_;
-#         for (@elem) { s/^/$INDENT/gm; }
-#         return "\n" . ($comment ? "$INDENT# $comment\n" : "") .
-#                join(",\n", @elem, "");
-    if ($comment || (@_ > $indent_lim && (length($tmp) > $LINEWIDTH || $tmp =~ /\n/))) {
-        my @lines = ( $comment ? "$INDENT# $comment" : () );
-        my @buf;
-        my $sep = $use_qw ? ' ' : ', ';
-        foreach my $v ( @_ )
+    my $tmp = "@$values";
+    my $sep = $use_qw ? ' ' : ', ';
+    if ($comment || (@$values > $indent_lim && (length($tmp) > $LINEWIDTH || $tmp =~ /\n/))) {
+        if( $use_qw )
         {
-            if( length( $INDENT . join( $sep, @buf, $v ) ) > $LINEWIDTH )
+            my @lines;
+            my @buf;
+            foreach my $v ( @$values )
             {
-                push( @lines, $INDENT . join( $sep, @buf ) );
-                @buf = ( $v );
-            }
-            else
-            {
-                push( @buf, $v );
+                if( scalar( @buf ) && length( $INDENT . join( ' ', @buf, $v ) ) > $LINEWIDTH )
+                {
+                    push( @lines, $INDENT . join( ' ', @buf ) );
+                    @buf = ( $v );
+                }
+                else
+                {
+                    push( @buf, $v );
+                }
             }
+            push( @lines, $INDENT . join( ' ', @buf ) ) if( scalar( @buf ) );
+            return (
+                $comment
+                    ? ( scalar( @lines ) > 1 ? "\n$INDENT" : '' ) . "# $comment" . ( scalar( @lines ) > 1 ? "\n" : '' )
+                    : ''
+            ) . 'qw(' . "\n" . join("\n", @lines,"") . ')';
+        }
+        else
+        {
+            my @elem = @$values;
+            for (@elem) { s/^/$INDENT/gm; }
+            return "\n" . ($comment ? "$INDENT# $comment\n" : "") .
+                   join(",\n", @elem, "");
         }
-        push( @lines, $INDENT . join( $sep, @buf ) ) if( scalar( @buf ) );
-        return ( $use_qw ? 'qw(' : '' ) . "\n" . join("\n", @lines,"") . ( $use_qw ? ')' : '' );
     } else {
-        return $use_qw ? 'qw( ' . join( ' ', @_ ) . ' )' : join(", ", @_);
+        return $use_qw ? 'qw( ' . join( $sep, @$values ) . ' )' : join($sep, @$values);
     }
 }
 
@@ -338,8 +345,9 @@ sub _dump
     my $ref  = ref $_[0];
     my $rval = $ref ? $_[0] : \$_[0];
     shift;
+    my $opts = {@_};
 
-    my($name, $idx, $dont_remember, $pclass, $pidx) = @_;
+    my($name, $idx, $dont_remember, $pclass, $pidx) = @$opts{qw( name idx dont_remember pclass pidx )};
 
     my($class, $type, $id);
     my $strval = overload::StrVal($rval);
@@ -358,7 +366,7 @@ sub _dump
     if ($] < 5.008 && $type eq "SCALAR") {
         $type = "REF" if $ref eq "REF";
     }
-    warn "\$$name(@$idx) ", ( $class || 'undef' ), " $type $id ($ref)" if $DEBUG;
+    warn "\$$name(@$idx) ", ( $class || 'undef' ), " $type $id ($ref), strval=$strval" if $DEBUG;
 
     my $out;
     my $comment;
@@ -372,7 +380,12 @@ sub _dump
             if (my $f = $filter->($ctx, $rval)) {
                 if (my $v = $f->{object}) {
                     local @FILTERS;
-                    $out = _dump($v, $name, $idx, 1);
+                    $out = _dump(
+                        $v,
+                        name => $name,
+                        idx => $idx,
+                        dont_remember => 1,
+                    );
                     $dont_remember++;
                 }
                 if (defined(my $c = $f->{bless})) {
@@ -463,7 +476,14 @@ sub _dump
             }
             else {
                 delete $seen{$id} if $type eq "SCALAR";  # will be seen again shortly
-                my $val = _dump($$rval, $name, [@$idx, "\$"], 0, $pclass, $pidx);
+                my $val = _dump(
+                    $$rval,
+                    name => $name,
+                    idx => [@$idx, "\$"],
+                    dont_remember => 0,
+                    pclass => $pclass,
+                    pidx => $pidx,
+                );
                 $out = $class ? "do{\\(my \$o = $val)}" : "\\$val";
             }
         # NOTE; regular string
@@ -493,7 +513,14 @@ sub _dump
     elsif ($type eq "GLOB") {
         if ($ref) {
             delete $seen{$id};
-            my $val = _dump($$rval, $name, [@$idx, "*"], 0, $pclass, $pidx);
+            my $val = _dump(
+                $$rval,
+                name => $name,
+                idx => [@$idx, "*"],
+                dont_remember => 0,
+                pclass => $pclass,
+                pidx => $pidx,
+            );
             $out = "\\$val";
             if ($out =~ /^\\\*Symbol::/) {
                 $require{Symbol}++;
@@ -509,7 +536,14 @@ sub _dump
                 next if $k eq "SCALAR" && ! defined $$gval;  # always there
                 my $f = scalar @fixup;
                 push(@fixup, "RESERVED");  # overwritten after _dump() below
-                $gval = _dump($gval, $name, [@$idx, "*{$k}"], 0, $pclass, $pidx);
+                $gval = _dump(
+                    $gval,
+                    name => $name,
+                    idx => [@$idx, "*{$k}"],
+                    dont_remember => 0,
+                    pclass => $pclass,
+                    pidx => $pidx,
+                );
                 $refcnt{$name}++;
                 my $gname = fullname($name, $idx);
                 $fixup[$f] = "$gname = $gval";  #XXX indent $gval
@@ -522,37 +556,27 @@ sub _dump
         my $tied = tied_str(tied(@$rval));
         # Quick check if we are dealing with a simple array of words/terms
         # and thus if we can use qw( .... ) instead of ( "some", "thing", "else" )
-#         my $use_qw = 1;
-#         my $only_numbers = 0;
-#         foreach my $v ( @$rval )
-#         {
-#             if( !defined( $v ) || 
-#                 ref( $v ) || 
-#                 # See perlop/"qw/STRING/" section
-#                 ( !ref( $v ) && $v =~ /[\,\#[:blank:]\h\v]/ ) )
-#             {
-#                 $use_qw = 0;
-#                 last;
-#             }
-#             $only_numbers++ if( $v =~ /^[-+]?(?:0|[1-9]\d*)(?:\.\d+)?\z/ );
-#         }
-#         # Don't use qw() if we are only dealing with numbers
-#         $use_qw = 0 if( $only_numbers == scalar( @$rval ) );
-#         
-#         if( $use_qw )
-#         {
-#             @vals = @$rval;
-#         }
-#         else
-#         {
-            my $i = 0;
-            for my $v (@$rval) {
-                push(@vals, _dump($v, $name, [@$idx, "[$i]"], $tied, $pclass, $pidx));
-                $i++;
-            }
-#         }
-#         $out = "[" . format_list(1, $tied, $use_qw, @vals) . "]";
-        $out = "[" . format_list(1, $tied, @vals) . "]";
+        my $use_qw = &_use_qw( $rval );
+
+        my $i = 0;
+        for my $v (@$rval) {
+            push(@vals, _dump(
+                $v,
+                name => $name,
+                idx => [@$idx, "[$i]"],
+                dont_remember => $tied,
+                pclass => $pclass,
+                pidx => $pidx,
+                use_qw => $use_qw,
+            ));
+            $i++;
+        }
+        $out = "[" . format_list(
+            paren => 1,
+            comment => $tied,
+            values => \@vals,
+            use_qw => $use_qw,
+        ) . "]";
     }
     # NOTE: hash
     elsif ($type eq "HASH") {
@@ -602,7 +626,14 @@ sub _dump
 
             push(@keys, $key);
             # push(@vals, _dump($$val, $name, [@$idx, "{$key}"], $tied, $pclass, $pidx));
-            my $this = _dump($$val, $name, [@$idx, "{$key}"], $tied, $pclass, $pidx);
+            my $this = _dump(
+                $$val,
+                name => $name,
+                idx => [@$idx, "{$key}"],
+                dont_remember => $tied,
+                pclass => $pclass,
+                pidx => $pidx,
+            );
             my $this_type;
             if ((my $i = index(overload::StrVal($$val), "(0x")) >= 0) {
                 $this_type = substr(overload::StrVal($$val), 0, $i);
@@ -636,7 +667,22 @@ sub _dump
     }
     # NOTE: code
     elsif ($type eq "CODE") {
-        $out = 'sub { ... }';
+        if( $CODE_DEPARSE && eval { require B::Deparse } )
+        {
+            # -sC to cuddle elsif, else and continue
+            # -si4 indent by 4 spaces (default)
+            # -p use extra parenthesis
+            # my $deparse = B::Deparse->new("-p", "-sC");
+            my $deparse = B::Deparse->new;
+            my $code = $deparse->coderef2text( $rval );
+            # Don't let our environment influence the code
+            1 while $code =~ s/^\{[\s\n]+use\s(warnings|strict);\n/\{\n/gs;
+            $out = 'sub ' . $code;
+        }
+        else
+        {
+            $out = 'sub { ... }';
+        }
     }
     # NOTE: vstring
     elsif ($type eq "VSTRING") {
@@ -659,6 +705,31 @@ sub _dump
     return $out;
 }
 
+sub _use_qw
+{
+    my $rval = shift( @_ );
+    # Quick check if we are dealing with a simple array of words/terms
+    # and thus if we can use qw( .... ) instead of ( "some", "thing", "else" )
+    my $use_qw = 1;
+    my $only_numbers = 0;
+    foreach my $v ( @$rval )
+    {
+        if( !defined( $v ) || 
+            ref( $v ) || 
+            substr( overload::StrVal( \$v ), 0, 7 ) eq 'VSTRING' ||
+            # See perlop/"qw/STRING/" section
+            ( !ref( $v ) && $v =~ /[\,\#[:blank:]\h\v]/ ) )
+        {
+            $use_qw = 0;
+            last;
+        }
+        $only_numbers++ if( $v =~ /^[-+]?(?:0|[1-9]\d*)(?:\.\d+)?\z/ );
+    }
+    # Don't use qw() if we are only dealing with numbers
+    $use_qw = 0 if( $only_numbers == scalar( @$rval ) || scalar( @$rval ) == 1 );
+    return( $use_qw );
+}
+
 1;
 # NOTE: POD
 __END__
@@ -680,7 +751,7 @@ Data::Pretty - Data Dump Beautifier
 
 =head1 VERSION
 
-    v0.1.2
+    v0.1.3
 
 =head1 DESCRIPTION
 
@@ -867,7 +938,7 @@ There are a few global variables that can be set to modify the output generated
 
 =head2 $Data::Pretty::INDENT
 
-This holds the string that's used for indenting multiline data structures. It's default value is C<"  "> (two spaces). Set it to C<""> to suppress indentation. Setting it to C<"| "> makes for nice visuals even if the dump output then fails to be valid Perl.
+This holds the string that's used for indenting multiline data structures. It's default value is C<"    "> (four spaces). Set it to C<""> to suppress indentation. Setting it to C<"| "> makes for nice visuals even if the dump output then fails to be valid Perl.
 
 =head2 $Data::Pretty::SHOW_UTF8
 
@@ -905,7 +976,7 @@ L<Data::Dump>, L<Data::Dumper>
 
 =head1 CREDITS
 
-Credits to Gisle Aas for the initial version and to Breno G. de Oliveira for maintaining it.
+Credits to Gisle Aas for the original L<Data::Dump> version and to Breno G. de Oliveira for maintaining it.
 
 =head1 COPYRIGHT & LICENSE
 
